@@ -7,13 +7,17 @@
     `(progn
       (export ',name :sb-posix)
       (defclass ,name ,superclasses
+         ;; KLUDGE: Splice out some slot options (they're
+         ;; for the conversion functions, not for DEFCLASS).
         ,(loop for slotd in slots
-               ;; KLUDGE: Splice out :ARRAY-LENGTH options (they're
-               ;; for the conversion functions, not for DEFCLASS).
-               for array-length-option = (member :array-length slotd)
-               collect (append (ldiff slotd array-length-option)
-                               (cddr array-length-option)))
+               collect
+               (let ((slotd (copy-list slotd)))
+                 (dolist (keyword '(:array-length :from-alien) slotd)
+                   (remf (cdr slotd) keyword))))
         ,@options)
+      ;; TODO (maybe): there's no reason to define to-alien routines
+      ;; struct stat, passwd, or group: OS interfaces only ever write
+      ;; into them, never read from them.
       (declaim (inline ,to-alien ,to-protocol))
       (declaim (inline ,to-protocol ,to-alien))
       (defun ,to-protocol (alien &optional instance)
@@ -26,6 +30,7 @@
                 ;;
                 ;; FIXME: baroque construction of intricate fragility
                 for array-length = (getf (cdr slotd) :array-length)
+                for from-alien = (getf (cdr slotd) :from-alien)
                 if array-length
                   collect `(progn
                              (let ((array (make-array ,array-length)))
@@ -37,8 +42,14 @@
                                         (sb-alien:slot alien ',(car slotd))
                                         i)))))
                 else
-                  collect `(setf (slot-value instance ',(car slotd))
-                                 (sb-alien:slot alien ',(car slotd))))
+                  collect (if from-alien
+                              ;; FROM-ALIEN is for any ad-hoc conversions that
+                              ;; SB-ALIEN doesn't automatically handle, such as
+                              ;; char** to list of strings.
+                              `(setf (slot-value instance ',(car slotd))
+                                     (,from-alien (sb-alien:slot alien ',(car slotd))))
+                              `(setf (slot-value instance ',(car slotd))
+                                     (sb-alien:slot alien ',(car slotd)))))
         instance)
       (defun ,to-alien (instance &optional alien)
         (declare (type (or null (sb-alien:alien (* ,alien-type))) alien)
@@ -55,6 +66,8 @@
                                         (sb-alien:slot alien ',(car slotd))
                                         i)
                                        (aref array i)))))
+                ;; N.B., nothing turns out to need a :TO-ALIEN
+                ;; counterpart of :FROM-ALIEN so far.
                 else
                   collect `(setf (sb-alien:slot alien ',(car slotd))
                                  (slot-value instance ',(car slotd)))))
@@ -147,17 +160,25 @@
 (define-call* "write" ssize-t minusp
   (fd file-descriptor) (buf (* t)) (count size-t))
 
-;;; FIXME: to detect errors in readdir errno needs to be set to 0 and
-;;; then checked, like it's done in sb-unix:readdir.
+(declaim (inline null-alien-and-errno-plusp))
+(defun null-alien-and-errno-plusp (alien)
+  (and (null-alien alien) (plusp (get-errno))))
+
+;; Slight wart: READDIR returns a null alien at the end of the
+;; directory; most other SB-POSIX interfaces (e.g., GETENV, GETPWNAM,
+;; GETGRGID, etc) return NIL instead of a null pointer for a non-error
+;; return. This might be the only detail in SB-POSIX that makes a user
+;; reach for SB-ALIEN, but it'd be an incompatible change to do
+;; anything about it.
 #+inode64
 (define-call ("readdir" :c-name "readdir$INODE64" :options :largefile)
   (* dirent)
-  not
+  null-alien-and-errno-plusp
   (dir (* t)))
 #-inode64
 (define-call (#-netbsd "readdir" #+netbsd "_readdir" :options :largefile)
   (* dirent)
-  not
+  null-alien-and-errno-plusp
   (dir (* t)))
 (define-call "closedir" int minusp (dir (* t)))
 ;; need to do this here because we can't do it in the DEFPACKAGE
@@ -335,8 +356,6 @@
 
   ;; FIXME this is a lie, of course this can fail, but there's no
   ;; error handling here yet!
-  #+darwin
-  (define-call "darwin_reinit" void never-fails)
   (define-call ("posix_fork" :c-name "fork") pid-t minusp)
   (defun fork ()
     "Forks the current process, returning 0 in the new process and the PID of
@@ -369,7 +388,8 @@ not supported."
       (sb-impl::finalizer-thread-start)
       (error "Cannot fork with multiple threads running."))
     (let ((pid (posix-fork)))
-      #+darwin (when (= pid 0) (darwin-reinit))
+      (when (= pid 0) ; child
+        (alien-funcall (extern-alien "sb_posix_after_fork" (function void))))
       #+sb-thread (sb-impl::finalizer-thread-start)
       pid))
   (export 'fork :sb-posix)
@@ -544,33 +564,147 @@ not supported."
 ;;; group database
 #-(or android win32)
 (define-protocol-class group alien-group ()
-  ((name :initarg :name :accessor group-name)
-   (passwd :initarg :passwd :accessor group-passwd)
-   (gid :initarg :gid :accessor group-gid)))
+  ((name :initarg :name :accessor group-name
+         :documentation "The name of the group.")
+   ;; Note: SUSv4 doesn't require this member
+   (passwd :initarg :passwd :accessor group-passwd
+           :documentation "The group's encrypted password.")
+   (gid :initarg :gid :accessor group-gid
+        :documentation "Numerical group ID.")
+   (mem :initarg :mem :accessor group-mem
+        ;; N.B., omitting any :TO-ALIEN because no alien interface
+        ;; reads from a filled-in struct group.
+        :from-alien sb-int:c-strings->string-list
+        :documentation "A list of strings naming members of the group."))
+  (:documentation
+   "Instances of this class represent entries in the system's group database."))
+
+;; None of the standardized interfaces to the user or group database
+;; is thread-safe or reentrant, at least with respect to getpwent or
+;; getgrent. The following two macros are for users to wrap around all
+;; getpw* or getgr* calls if they want to do things safely. (Note that
+;; on #-sb-thread, this still protects against reentrancy, but doesn't
+;; grab a lock.)
+#-(or android win32)
+(macrolet ((define-database-protection-form (dbname)
+  (let* ((macro-name (intern (format nil "WITH-~A-DATABASE" dbname) :sb-posix))
+         (lock-var (intern (format nil "*~A-DATABASE-LOCK*" dbname) :sb-posix))
+         (lock-name (format nil "~A database lock" dbname))
+         (special-var (intern (format nil "*WITH-~A-DATABASE*" dbname) :sb-posix))
+         (reentrancy-error-message
+          (format nil "~@(~A~) database access is not reentrant." dbname))
+         (assertion (intern (format nil "ASSERT-~A" macro-name) :sb-posix)))
+    #-sb-thread (declare (ignore lock-var lock-name))
+    `(progn
+       #+sb-thread
+       (sb-ext:defglobal ,lock-var
+        (sb-thread:make-mutex :name ,lock-name))
+       (defvar ,special-var nil)
+       (defmacro ,macro-name (&body body)
+         `(progn
+            (when ,',special-var
+              (error ,',reentrancy-error-message))
+            #+sb-thread
+            (sb-thread:with-mutex (,',lock-var)
+             (let ((,',special-var t))
+               ,@body))
+            #-sb-thread
+            (let ((,',special-var t))
+              ,@body)))
+       (defmacro ,assertion (function)
+         `(unless ,',special-var
+            (error "~A may only be called during ~A."
+                    ',function ',',macro-name)))))))
+(define-database-protection-form passwd)
+(define-database-protection-form group))
 
 #-(or win32 android)
-(macrolet ((define-obj-call (name arg type conv)
+(macrolet ((define-obj-call (name result-type conv with-macro assertion
+                             &optional arg arg-type)
   ;; FIXME: this isn't the documented way of doing this, surely?
   (let ((lisp-name (intern (string-upcase name) :sb-posix)))
     `(progn
-      (export ',lisp-name :sb-posix)
+      ;; Don't export GET/SET/END/??ENT bindings; the iteration
+      ;; macros are less error-prone.
+      ,@(when arg `((export ',lisp-name :sb-posix)))
       (declaim (inline ,lisp-name))
-      (defun ,lisp-name (,arg)
-        (let ((r (alien-funcall (extern-alien ,name ,type) ,arg)))
+      (defun ,lisp-name (,@(when arg `(,arg)))
+        ,@(unless arg
+            `((,assertion ',lisp-name)))
+        (set-errno 0)
+        (let ((r (,@(if arg `(,with-macro) '(progn))
+                    (alien-funcall
+                     (extern-alien
+                      ,name (function ,result-type ,@(when arg-type `(,arg-type))))
+                   ,@(when arg `(,arg))))))
           (if (null-alien r)
-              nil
-              (,conv r))))))))
+              (when (plusp (get-errno))
+                (syscall-error ',lisp-name))
+              (,conv r)))))))
+  (define-enumerator-call (name assertion)
+      (let ((lisp-name (intern (string-upcase name) :sb-posix)))
+        `(progn
+           (declaim (inline ,lisp-name))
+           (defun ,lisp-name ()
+             (,assertion ',lisp-name)
+             (alien-funcall (extern-alien ,name (function void))))))))
 
-(define-obj-call "getpwnam" login-name (function (* alien-passwd) (c-string :not-null t))
-                 alien-to-passwd)
-(define-obj-call "getpwuid" uid (function (* alien-passwd) uid-t)
-                 alien-to-passwd)
-(define-obj-call "getgrnam" login-name (function (* alien-group) (c-string :not-null t))
-                 alien-to-group)
-(define-obj-call "getgrgid" gid (function (* alien-group) gid-t)
-                 alien-to-group)
+;; passwd database
+(define-obj-call "getpwnam" (* alien-passwd) alien-to-passwd
+                 with-passwd-database assert-with-passwd-database
+                 login-name (c-string :not-null t))
+(define-obj-call "getpwuid" (* alien-passwd) alien-to-passwd
+                 with-passwd-database assert-with-passwd-database
+                 uid uid-t)
+(define-obj-call "getpwent" (* alien-passwd) alien-to-passwd
+                 with-passwd-database assert-with-passwd-database)
+;; Including these here for thematic grouping.
+(define-enumerator-call "setpwent" assert-with-passwd-database)
+(define-enumerator-call "endpwent" assert-with-passwd-database)
+
+;; Same thing, but for the group database.
+(define-obj-call "getgrnam" (* alien-group) alien-to-group
+                 with-group-database assert-with-group-database
+                 group-name (c-string :not-null t))
+(define-obj-call "getgrgid" (* alien-group) alien-to-group
+                 with-group-database assert-with-group-database
+                 gid gid-t)
+(define-obj-call "getgrent" (* alien-group) alien-to-group
+                 with-group-database assert-with-group-database)
+(define-enumerator-call "setgrent" assert-with-group-database)
+(define-enumerator-call "endgrent" assert-with-group-database)
 ) ; end MACROLET
 
+#-(or android win32)
+(macrolet ((define-database-iterator (dbname with set get end)
+  (let* ((name (intern (format nil "DO-~AS" dbname) :sb-posix))
+         (docstring
+          (let ((*print-right-margin* 70))
+            (format
+             nil
+             "~@<Evaluate BODY with ~A bound to successive entries from ~
+              the ~:*~(~A~) database, and return RESULT. ~
+              An implicit block named NIL surrounds the form; ~
+              an implicit TAGBODY surrounds BODY. ~
+              It is unspecified whether ~:*~A is assigned, rebound, or ~
+              destructively modified upon each iteration. ~
+              It is an error to use any operator that accesses the ~
+              ~0@*~A database during the dynamic extent of ~A.~:@>"
+             dbname name))))
+    `(progn
+       (export ',name :sb-posix)
+       (defmacro ,name ((,dbname &optional result) &body body)
+         ,docstring
+         `(,',with
+           (,',set)
+           (do ((,,dbname (,',get) (,',get)))
+               ((null ,,dbname) (,',end) ,result)
+             ,@body)))))))
+  (define-database-iterator passwd with-passwd-database
+                            setpwent getpwent endpwent)
+  (define-database-iterator group with-group-database
+                            setgrent getgrent endgrent))
+
 #-win32
 (define-protocol-class timeval alien-timeval ()
   ((sec :initarg :tv-sec :accessor timeval-sec
@@ -782,10 +916,9 @@ not supported."
           result)))
   (export 'utime :sb-posix)
   (defun utime (filename &optional access-time modification-time)
-    (let ((fun (extern-alien #-netbsd "utime" #+netbsd "_utime"
-                             (function int (c-string :not-null t)
-                                       (* alien-utimbuf))))
-          (name (filename filename)))
+  (with-alien ((fun (function int (c-string :not-null t) (* alien-utimbuf))
+                    :extern #-netbsd "utime" #+netbsd "_utime"))
+    (let ((name (filename filename)))
       (if (not (and access-time modification-time))
           (alien-funcall fun name nil)
           (with-alien ((utimbuf (struct alien-utimbuf)))
@@ -794,7 +927,7 @@ not supported."
             (let ((result (alien-funcall fun name (alien-sap utimbuf))))
               (if (minusp result)
                   (syscall-error 'utime)
-                  result))))))
+                  result)))))))
   (export 'utimes :sb-posix)
   (defun utimes (filename &optional access-time modification-time)
     (flet ((seconds-and-useconds (time)
@@ -805,10 +938,9 @@ not supported."
              (if (minusp value)
                  (syscall-error 'utimes)
                  value)))
-      (let ((fun (extern-alien #-netbsd "utimes" #+netbsd "sb_utimes"
-                               (function int (c-string :not-null t)
-                                                  (* (array alien-timeval 2)))))
-            (name (filename filename)))
+   (with-alien ((fun (function int (c-string :not-null t) (* (array alien-timeval 2)))
+                     :extern #-netbsd "utimes" #+netbsd "sb_utimes"))
+     (let ((name (filename filename)))
         (if (not (and access-time modification-time))
             (maybe-syscall-error (alien-funcall fun name nil))
             (with-alien ((buf (array alien-timeval 2)))
@@ -821,18 +953,22 @@ not supported."
                               (slot modtime 'usec))
                       (seconds-and-useconds (or modification-time 0)))
                 (maybe-syscall-error (alien-funcall fun name
-                                                    (alien-sap buf))))))))))
+                                                    (alien-sap buf)))))))))))
 
 
 ;;; environment
 
 (defun getenv (name)
+  ;; SUSv4 doesn't define any errors for getenv, but some systems do.
+  (set-errno 0)
   (let ((r (alien-funcall
             (extern-alien "getenv" (function (* char) (c-string :not-null t)))
             name)))
     (declare (type (alien (* char)) r))
-    (unless (null-alien r)
-      (cast r c-string))))
+    (if (null-alien r)
+        (when (plusp (get-errno))
+          (syscall-error 'getenv))
+        (cast r c-string))))
 #-win32
 (progn
   (define-call "setenv" int minusp
