@@ -748,33 +748,34 @@ specifies to signal a warning if SWANK package is in variance, and an error othe
     (briefly-describe-symbol symbol))
   (values))
 
-(defun perfectly-hash-package (designator)
+;;; TODO: I think the original idea was to place all symbols into a single
+;;; vector, with an bit-vector to record internal/external per symbol.
+(defun perfectly-hash-package (package)
   (flet ((rehash (table &aux (cells (symtbl-%cells table)))
            (when (functionp (car cells))
              (return-from rehash)) ; already hashed
+           ;; The APROPOS-LIST R/O scan optimization is inadmissible if no R/O space
+           #-darwin-jit (setf (symtbl-modified table) nil)
            (let* ((cells (cdr cells))
                   (hashes (map '(simple-array (unsigned-byte 32) (*))
-                               (lambda (symbol) (ldb (byte 32 0) (symbol-hash symbol)))
-                               (remove-if-not #'symbolp cells))))
-             (when (or (< (length (remove-duplicates hashes)) (length hashes))
-                       ;; limitation of perfecthash, also probably no benefit
-                       (< (length hashes) 5))
+                               #'symbol-name-hash
+                               (remove-if-not #'symbolp cells)))
+                  (hash-expr (sb-c:make-perfect-hash-lambda hashes)))
+             (unless hash-expr
                (return-from rehash))
-             (let* ((hash-expr (sb-c:make-perfect-hash-lambda hashes))
-                    (fun (compile nil hash-expr))
-                    (new-cells (make-array (length hashes) :initial-element 0)))
+             (let ((fun (compile nil hash-expr))
+                   (new-cells (make-array (length hashes) :initial-element 0)))
                (dovector (s cells)
                  (when (symbolp s)
-                   (let ((hash (funcall fun (ldb (byte 32 0) (symbol-hash s)))))
+                   (let ((hash (funcall fun (symbol-name-hash s))))
                      (aver (eq (svref new-cells hash) 0))
                      (setf (svref new-cells hash) s))))
                (setf (symtbl-%cells table) (cons fun new-cells)
                      (symtbl-size table) (length hashes)
                      (symtbl-free table) 0
                      (symtbl-deleted table) 0)))))
-    (let ((package (find-package designator)))
-      (rehash (package-internal-symbols package))
-      (rehash (package-external-symbols package)))))
+    (rehash (package-internal-symbols package))
+    (rehash (package-external-symbols package))))
 
 ;;; Reorganize both hashsets of all packages, called by SAVE-LISP-AND-DIE.
 ;;; With a few exceptions, tables are saved at 100% load factor and a perfect hash.
@@ -797,14 +798,18 @@ specifies to signal a warning if SWANK package is in variance, and an error othe
 ;;;
 (defun tune-hashset-sizes-of-all-packages ()
   (flet ((tune-table-size (desired-lf table)
-           (resize-symbol-table table (%symtbl-count table) t desired-lf)
+           (resize-symbol-table table (%symtbl-count table) 'intern desired-lf)
            ;; The APROPOS-LIST R/O scan optimization is inadmissible if no R/O space
            #-darwin-jit (setf (symtbl-modified table) nil)))
     (dolist (package (list-all-packages))
       ;; Choose load factor based on whether INTERN is expected at runtime
+      ;; FIXME: because changing a package from perfectly hashed back to
+      ;; an open-addressing table is not thread-safe, _only_ the CL package can
+      ;; become perfectly hashed.
       (let ((lf (cond ((eq (the package package) *keyword-package*) 60/100)
                       ((eq (package-id package) +package-id-user+) 8/10)
-                      (t 1))))
+                      ((eq package *cl-package*) 1)
+                      (t 8/10))))
         (cond ((= lf 1)
                (perfectly-hash-package package))
               (t
@@ -822,7 +827,7 @@ specifies to signal a warning if SWANK package is in variance, and an error othe
            (when (functionp reciprocals)
              (return-from metrics (values 1 1 1))) ; 1 probe max+avg, 100% load
            (flet ((probe-seq-len (symbol)
-                    (let* ((name-hash (sxhash symbol))
+                    (let* ((name-hash (symbol-name-hash symbol))
                            (index (symbol-table-hash 1 name-hash nslots))
                            (h2 (symbol-table-hash 2 name-hash nslots))
                            (nprobes 1))
