@@ -63,15 +63,6 @@
                      :drop-through drop-through
                      :immediate-tested immediate-tested)))))
 
-;; I can see no reason this would ever be used.
-;; (or fixnum character|unbound-marker) is implausible.
-(defun %test-fixnum-and-immediate (value target not-p immediate)
-  (error "WAT")
-  (let ((drop-through (gen-label)))
-    (generate-fixnum-test value)
-    (inst jmp :z (if not-p drop-through target))
-    (%test-immediate value target not-p immediate drop-through)))
-
 ;; Numerics
 (defun %test-fixnum-immediate-and-headers (value temp target not-p immediate headers
                                            &key value-tn-ref immediate-tested)
@@ -100,11 +91,10 @@
                                      :immediate-tested immediate-tested)))))
 
 (defun %test-immediate (value temp target not-p immediate
-                        &optional (drop-through (gen-label)))
-  (declare (ignore temp))
+                        &key value-tn-ref)
+  (declare (ignore temp value-tn-ref))
   (inst cmp :byte value immediate)
-  (inst jmp (if not-p :ne :e) target)
-  (emit-label drop-through))
+  (inst jmp (if not-p :ne :e) target))
 
 ;; Numerics including short-float, excluding fixnum
 (defun %test-immediate-and-headers (value temp target not-p immediate headers
@@ -123,7 +113,8 @@
                  :value-tn-ref value-tn-ref
                  :immediate-tested immediate-tested))
 
-(defun %test-lowtag (value temp target not-p lowtag)
+(defun %test-lowtag (value temp target not-p lowtag &key value-tn-ref)
+  (declare (ignore value-tn-ref))
   (%lea-for-lowtag-test temp value lowtag)
   (inst test :byte temp lowtag-mask)
   (inst jmp (if not-p :nz :z) target))
@@ -146,7 +137,10 @@
                               (not except)
                               (or (atom (car headers))
                                   (= (caar headers) bignum-widetag)
-                                  (= (cdar headers) complex-array-widetag)))
+                                  (= (cdar headers) complex-array-widetag)
+                                  (and value-tn-ref
+                                       (= (caar headers) simple-array-widetag)
+                                       (csubtypep (tn-ref-type value-tn-ref) (specifier-type 'array)))))
                          (ea (- lowtag) value)
                          temp))
          (first (car headers))
@@ -219,12 +213,15 @@
           (let ((header (car remaining))
                 (last (null (cdr remaining))))
             (cond
-             ((atom header)
-              (inst cmp :byte widetag-tn header)
-              (if last
-                  (inst jmp equal target)
-                  (inst jmp :e when-true)))
-             (t
+              ((and (eql header simple-array-widetag)
+                    value-tn-ref
+                    (csubtypep (tn-ref-type value-tn-ref) (specifier-type 'string))))
+              ((atom header)
+               (inst cmp :byte widetag-tn header)
+               (if last
+                   (inst jmp equal target)
+                   (inst jmp :e when-true)))
+              (t
                (let ((start (car header))
                      (end (cdr header)))
                  (cond
@@ -338,42 +335,59 @@
     (inst test value value)))
 
 (macrolet ((define (name src-size)
-             `(progn
-                (define-vop (,name)
-                  (:translate ,name)
-                  (:args (value :scs (any-reg descriptor-reg)))
-                  (:conditional :z)
-                  (:arg-refs arg-ref)
-                  (:policy :fast-safe)
-                  (:temporary (:sc unsigned-reg) temp temp2)
-                  (:generator 6
-                    ;; Optimistically assume that the argument is a fixnum.
-                    ;; If it is, then we're shifting out the tag bit, otherwise
-                    ;; putting random bits in the low byte, which is harmless.
-                    (move temp value)
-                    (inst sar temp n-fixnum-tag-bits)
-                    (inst movsx '(,src-size :qword) temp2 temp)
-                    (inst cmp temp2 temp)
-                    ;; If the input wasn't known to be a fixnum, and it passed
-                    ;; the sign-extension test, then check FIXNUMP now.
-                    (unless (csubtypep (tn-ref-type arg-ref) (specifier-type 'fixnum))
-                      (inst jmp :nz OUT) ; did NOT pass the sign-extension test
-                      (inst test :byte value fixnum-tag-mask))
-                    OUT))
-                (define-vop (,(symbolicate name "/SIGNED"))
-                  (:translate ,name)
-                  (:args (value :scs (signed-reg)))
-                  (:arg-types signed-num)
-                  (:conditional :z)
-                  (:policy :fast-safe)
-                  (:temporary (:sc unsigned-reg) temp)
-                  (:generator 2
-                    (inst movsx '(,src-size :qword) temp value)
-                    (inst cmp temp value))))))
+             `(define-vop (,(symbolicate name "/SIGNED"))
+                (:translate ,name)
+                (:args (value :scs (signed-reg)))
+                (:arg-types signed-num)
+                (:conditional :z)
+                (:policy :fast-safe)
+                (:temporary (:sc unsigned-reg) temp)
+                (:generator 2
+                  (inst movsx '(,src-size :qword) temp value)
+                  (inst cmp temp value)))))
   (define signed-byte-8-p :byte)
   (define signed-byte-16-p :word)
   (define signed-byte-32-p :dword))
 
+(define-vop (signed-byte-8-p)
+  (:translate signed-byte-8-p)
+  (:args (value :scs (any-reg descriptor-reg)))
+  (:conditional :z)
+  (:arg-refs arg-ref)
+  (:policy :fast-safe)
+  (:temporary (:sc unsigned-reg) temp)
+  (:generator 6
+    (inst lea temp (ea (ash (expt 2 7) n-fixnum-tag-bits) value))
+    (inst test temp (lognot (fixnumize (1- (expt 2 8)))))))
+
+(define-vop (signed-byte-16-p)
+  (:translate signed-byte-16-p)
+  (:args (value :scs (any-reg descriptor-reg)))
+  (:conditional :z)
+  (:arg-refs arg-ref)
+  (:policy :fast-safe)
+  (:temporary (:sc unsigned-reg) temp)
+  (:generator 6
+    (inst lea temp (ea (ash (expt 2 15) n-fixnum-tag-bits) value))
+    (inst test temp (lognot (fixnumize (1- (expt 2 16)))))))
+
+(define-vop (signed-byte-32-p)
+  (:translate signed-byte-32-p)
+  (:args (value :scs (any-reg descriptor-reg)))
+  (:conditional :z)
+  (:arg-refs arg-ref)
+  (:policy :fast-safe)
+  (:temporary (:sc unsigned-reg) temp temp2)
+  (:generator 6
+    (move temp value)
+    (inst sar temp n-fixnum-tag-bits)
+    (inst movsx '(:dword :qword) temp2 temp)
+    (inst cmp temp2 temp)
+    (unless (csubtypep (tn-ref-type arg-ref) (specifier-type 'fixnum))
+      (inst jmp :nz out)
+      (inst test :byte value fixnum-tag-mask))
+    out))
+
 ;;; Sign bit and fixnum tag bit.
 (defconstant non-negative-fixnum-mask-constant
   #x8000000000000001)
@@ -452,65 +466,62 @@
 
         (emit-label not-target)))))
 
-(defun power-of-two-limit-p (x)
-  (and (fixnump x)
-       (= (logcount (1+ x)) 1)))
-
-(define-vop (test-fixnum-mod-power-of-two)
-  (:args (value :scs (any-reg descriptor-reg
-                              unsigned-reg signed-reg)))
-  (:arg-types *
-              (:constant (satisfies power-of-two-limit-p)))
-  (:translate fixnum-mod-p)
-  (:conditional :e)
-  (:info hi)
-  (:policy :fast-safe)
-  (:arg-refs value-ref)
-  (:vop-var vop)
-  (:generator 4
-    (let ((fixnum-hi (if (sc-is value unsigned-reg signed-reg)
-                         hi
-                         (fixnumize hi))))
-      (cond ((sc-is value descriptor-reg)
-             (let* ((mask (lognot fixnum-hi))
-                    (constant
-                      (if (= (ldb (byte 64 0) mask) non-negative-fixnum-mask-constant)
-                          (ea non-negative-fixnum-mask-constant-wired-address)
-                          (constantize mask))))
-               (inst test value constant)))
-            (t
-             (change-vop-flags vop '(:be))
-             (inst cmp value (constantize fixnum-hi)))))))
-
-(define-vop (test-fixnum-mod-tagged-unsigned)
-  (:args (value :scs (any-reg unsigned-reg signed-reg)))
-  (:arg-types (:or tagged-num unsigned-num signed-num)
-              (:constant fixnum))
-  (:translate fixnum-mod-p)
-  (:conditional :be)
-  (:info hi)
-  (:policy :fast-safe)
-  (:generator 5
-     (let ((fixnum-hi (if (sc-is value unsigned-reg signed-reg)
-                          hi
-                          (fixnumize hi))))
-       (inst cmp value (constantize fixnum-hi)))))
-
-(define-vop (test-fixnum-mod-*)
-  (:args (value :scs (any-reg descriptor-reg)))
-  (:arg-types * (:constant fixnum))
-  (:translate fixnum-mod-p)
-  (:conditional)
-  (:info target not-p hi)
-  (:policy :fast-safe)
-  (:generator 6
-     (let* ((fixnum-hi (fixnumize hi))
-            (skip (gen-label)))
-       (generate-fixnum-test value)
-       (inst jmp :ne (if not-p target skip))
-       (inst cmp value (constantize fixnum-hi))
-       (inst jmp (if not-p :a :be) target)
-       (emit-label skip))))
+(define-vop (unsigned-byte-x-p type-predicate)
+  (:arg-types * (:constant (integer 1)))
+  (:translate sb-c::unsigned-byte-x-p)
+  (:info target not-p x)
+  (:temporary (:sc unsigned-reg) last-digit)
+  (:generator 10
+    (let* ((type (tn-ref-type args))
+           (fixnum-p (types-equal-or-intersect type (specifier-type 'fixnum)))
+           (integer-p (csubtypep type (specifier-type 'integer)))
+           (unsigned-p (not (types-equal-or-intersect type (specifier-type '(integer * -1))))))
+      (multiple-value-bind (yep nope)
+          (if not-p
+              (values not-target target)
+              (values target not-target))
+        (assemble ()
+          (when fixnum-p
+            (cond (unsigned-p
+                   (inst test :byte value fixnum-tag-mask)
+                   (inst jmp :z yep))
+                  (t ;; Is it a fixnum with the sign bit clear?
+                   (inst test (ea non-negative-fixnum-mask-constant-wired-address) value)
+                   (inst jmp :z yep))))
+          (cond ((fixnum-or-other-pointer-tn-ref-p args t)
+                 (when (and fixnum-p
+                            (not unsigned-p))
+                   (inst test :byte value fixnum-tag-mask)
+                   (inst jmp :z nope)))
+                (t
+                 (%lea-for-lowtag-test temp value other-pointer-lowtag)
+                 (inst test :byte temp lowtag-mask)
+                 (inst jmp :ne nope)))
+          ;; Get the header.
+          (loadw temp value 0 other-pointer-lowtag)
+          (unless integer-p
+            (inst cmp :byte temp bignum-widetag)
+            (inst jmp :ne nope))
+          (inst shr temp n-widetag-bits)
+          (inst cmp :dword temp (1+ (/ x n-word-bits)))
+          (inst jmp :g nope)
+          ;; Is it a sign-extended sign bit
+          (cond (unsigned-p
+                 (inst jmp :l yep)
+                 (inst cmp :dword (ea (+ (- other-pointer-lowtag) (/ n-word-bytes 2))
+                                      value temp n-word-bytes)
+                       0)
+                 (inst jmp (if not-p :nz :z) target))
+                (t
+                 (inst mov last-digit (ea (- other-pointer-lowtag) value temp n-word-bytes))
+                 (inst jmp :l fixnum)
+                 (inst test last-digit last-digit)
+                 (inst jmp :nz nope)))
+          fixnum
+          (unless unsigned-p
+            (inst test last-digit last-digit)
+            (inst jmp (if not-p :s :ns) target)))))
+    not-target))
 
 ;;; SINGLE-FLOAT-P, CHARACTERP, UNBOUND-MARKER-P produce a flag result
 ;;; and never need a temporary.
@@ -564,7 +575,7 @@
                   out))))
   (define bignump (bignum-widetag))
   (define ratiop (ratio-widetag))
-  (define complex-rational-p (complex-widetag))
+  (define complex-rational-p (complex-rational-widetag))
   (define complex-single-float-p (complex-single-float-widetag))
   (define complex-double-float-p (complex-double-float-widetag))
   (define double-float-p (double-float-widetag))
@@ -611,10 +622,17 @@
                     (:info)
                     (:conditional :c) ; Carry flag = "below" (unsigned)
                     (:arg-refs value-tn-ref)
+                    (:vop-var vop)
                     (:generator 4
-                      (fail-if-not-otherptr)
-                      (inst sub :byte temp ,min)
-                      (inst cmp :byte temp ,(1+ (- max min)))
+                      (cond ((and (eq ,min simple-array-widetag)
+                                  (csubtypep (tn-ref-type value-tn-ref) (specifier-type 'array)))
+                             (change-vop-flags vop '(:le))
+                             (inst cmp :byte (ea (- other-pointer-lowtag) value)
+                                   ,max))
+                            (t
+                             (fail-if-not-otherptr)
+                             (inst sub :byte temp ,min)
+                             (inst cmp :byte temp ,(1+ (- max min)))))
                       OUT)))))
     (define simple-rank-1-array-*-p +simple-rank-1-array-widetags+)
     (define vectorp +vector-widetags+)
@@ -688,18 +706,16 @@
   (:translate keywordp)
   (:generator 3
     (cond ((csubtypep (tn-ref-type args) (specifier-type 'symbol))
-           (inst cmp :word (ea (+ (ash symbol-name-slot word-shift) 6
-                                  (- other-pointer-lowtag))
-                               value)
+           (inst cmp :word (ea (- 1 other-pointer-lowtag) value)
                  sb-impl::+package-id-keyword+))
           (t
            (inst lea temp (ea (- other-pointer-lowtag) value))
            (inst test :byte temp lowtag-mask)
            (inst jmp :ne out)
-           (inst cmp :byte (ea temp) symbol-widetag)
-           (inst jmp :ne out)
-           (inst cmp :word (ea (+ (ash symbol-name-slot word-shift) 6) temp)
-                 sb-impl::+package-id-keyword+)))
+           (inst mov :dword temp (ea temp))
+           (inst shl :dword temp 8) ; zeroize flag/generation bits
+           (inst cmp :dword temp
+                 (ash (logior (ash sb-impl::+package-id-keyword+ 8) symbol-widetag) 8))))
     out))
 
 (define-vop (consp type-predicate)
@@ -740,10 +756,8 @@
           (value :scs (any-reg descriptor-reg)))
    (:vop-var vop)
    (:temporary (:sc unsigned-reg) temp)
-   (:node-var node)
    (:generator 1
-     (unless (sb-c::set-slot-old-p node)
-       (emit-gengc-barrier object nil temp (vop-nth-arg 1 vop) value))
+     (emit-gengc-barrier object nil temp (vop-nth-arg 1 vop))
      (inst mov :dword (ea (- 4 instance-pointer-lowtag) object) value)))
  (define-vop (%fun-layout %instance-layout)
    (:translate %fun-layout)
@@ -751,8 +765,7 @@
  (define-vop (%set-fun-layout %set-instance-layout)
    (:translate %set-fun-layout)
    (:generator 1
-     (unless (sb-c::set-slot-old-p node)
-       (emit-gengc-barrier object nil temp (vop-nth-arg 1 vop) value))
+     (emit-gengc-barrier object nil temp (vop-nth-arg 1 vop))
      (inst mov :dword (ea (- 4 fun-pointer-lowtag) object) value)))
  (define-vop ()
   (:translate sb-c::layout-eq)
@@ -797,7 +810,7 @@
 (define-vop (fixnump simple-type-predicate)
   (:translate fixnump)
   (:arg-refs arg-ref)
-  (:args (value :scs (any-reg descriptor-reg) :load-if (tn-ref-memory-access arg-ref)))
+  (:args (value :scs (any-reg descriptor-reg)))
   (:conditional :z)
   ;; the compiler is very sensitive to this cost here as regards boxing. DON'T TOUCH !!!
   (:generator 3
@@ -900,24 +913,42 @@
         (assemble ()
           (when (types-equal-or-intersect (tn-ref-type integer-ref) (specifier-type 'fixnum))
             (generate-fixnum-test integer)
-            (inst jmp :nz BIGNUM)
-            (cond ((eql fixnum 0)
-                   (inst test integer integer))
-                  ((or (and (eql fixnum #.(fixnumize -1))
-                            (or (and (eql comparison :le)
-                                     (setf comparison :l))
-                                (and (eql comparison :g)
-                                     (setf comparison :ge))))
-                       (and (eql fixnum #.(fixnumize 1))
-                            (or (and (eql comparison :ge)
-                                     (setf comparison :g))
-                                (and (eql comparison :l)
-                                     (setf comparison :le)))))
-                   (inst test integer integer))
-                  (t
-                   (inst cmp integer fixnum)))
-            (inst jmp comparison yep)
-            (inst jmp nope))
+            (cond
+              ((and (eql fixnum 0)
+                    (memq comparison '(:ge :l))
+                    (csubtypep (type-intersection (tn-ref-type integer-ref)
+                                                  (specifier-type 'fixnum))
+                               (specifier-type '(integer * -1))))
+               (inst jmp :z (if (eq comparison :l)
+                                yep
+                                nope)))
+              ((and (eql fixnum 0)
+                    (memq comparison '(:ge :l))
+                    (csubtypep (type-intersection (tn-ref-type integer-ref)
+                                                  (specifier-type 'fixnum))
+                               (specifier-type '(integer 0))))
+               (inst jmp :z (if (eq comparison :l)
+                                nope
+                                yep)))
+              (t
+               (inst jmp :nz BIGNUM)
+               (cond ((eql fixnum 0)
+                      (inst test integer integer))
+                     ((or (and (eql fixnum #.(fixnumize -1))
+                               (or (and (eql comparison :le)
+                                        (setf comparison :l))
+                                   (and (eql comparison :g)
+                                        (setf comparison :ge))))
+                          (and (eql fixnum #.(fixnumize 1))
+                               (or (and (eql comparison :ge)
+                                        (setf comparison :g))
+                                   (and (eql comparison :l)
+                                        (setf comparison :le)))))
+                      (inst test integer integer))
+                     (t
+                      (inst cmp integer fixnum)))
+               (inst jmp comparison yep)
+               (inst jmp nope))))
           bignum
           (unless (fixnum-or-other-pointer-tn-ref-p integer-ref t)
             (test-type integer temp nope t (other-pointer-lowtag)))
@@ -927,7 +958,7 @@
             (inst jmp :ne nope))
           #.(assert (= (integer-length bignum-widetag) 5))
           (inst shr temp 5)
-          (inst cmp :qword (ea (- other-pointer-lowtag) integer temp) 0)
+          (inst cmp :dword (ea (+ (- other-pointer-lowtag) (/ n-word-bytes 2)) integer temp) 0)
           (inst jmp (case comparison
                       ((:l :le) (if not-p :ge :l))
                       (t (if not-p :l :ge)))
@@ -966,19 +997,23 @@
 
 (define-vop (load-other-pointer-widetag)
   (:args (value :scs (any-reg descriptor-reg)))
-  (:arg-refs args)
+  (:arg-refs value-ref)
   (:info not-other-pointer-label null-label zero-extend)
   (:results (r :scs (unsigned-reg)))
   (:result-types unsigned-num)
   (:generator 1
-    (cond ((other-pointer-tn-ref-p args (not null-label))
+    (when null-label
+      (if (types-equal-or-intersect
+           (type-difference (tn-ref-type value-ref) (specifier-type 'null))
+           (specifier-type 'cons))
+          (inst cmp value nil-value)
+          (inst cmp :byte value (logand nil-value #xff)))
+      (inst jmp :e null-label))
+    (cond ((other-pointer-tn-ref-p value-ref t)
            (if zero-extend
                (inst movzx '(:byte :dword) r (ea (- other-pointer-lowtag) value))
                (inst mov :byte r (ea (- other-pointer-lowtag) value))))
           (t
-           (when null-label
-             (inst cmp value nil-value)
-             (inst jmp :e null-label))
            (%lea-for-lowtag-test r value other-pointer-lowtag :qword)
            (inst test :byte r lowtag-mask)
            (inst jmp :nz not-other-pointer-label)
