@@ -37,8 +37,7 @@
 ;;; be used.
 ;;; This is only enabled for certain build configurations that need
 ;;; a scratch register in various random places.
-#-ubsan (eval-when (:compile-toplevel :load-toplevel :execute)
-          (defconstant global-temp-reg nil))
+#-ubsan (defconstant global-temp-reg nil)
 #+ubsan (progn (define-symbol-macro temp-reg-tn r11-tn)
                (defconstant global-temp-reg 11))
 
@@ -89,11 +88,7 @@
 
 (macrolet ((defreg (name offset size)
              (declare (ignore size))
-             `(eval-when (:compile-toplevel :load-toplevel :execute)
-                    ;; EVAL-WHEN is necessary because stuff like #.EAX-OFFSET
-                    ;; (in the same file) depends on compile-time evaluation
-                    ;; of the DEFCONSTANT. -- AL 20010224
-                (defconstant ,(symbolicate name "-OFFSET") ,offset)))
+             `(defconstant ,(symbolicate name "-OFFSET") ,offset))
            (defregset (name &rest regs)
              ;; FIXME: this would be DEFCONSTANT-EQX were it not
              ;; for all the style-warnings about earmuffs on a constant.
@@ -108,9 +103,7 @@
            (define-gprs (want-offsets offsets-list names array)
              `(progn
                 (defconstant-eqx ,names ,array #'equalp)
-                ;; We need the constants evaluable because of the DEFGLOBAL
-                ;; which is needed because of forms such as #.*qword-regs*
-                (eval-when (:compile-toplevel :load-toplevel :execute)
+                (progn
                   ,@(when want-offsets
                       (let ((i -1))
                         (map 'list
@@ -467,8 +460,9 @@
      (when (or (static-symbol-p value)
                ;; The cross-compiler always uses immobile-space if it exists.
                #+(and immobile-space sb-xc-host) t
-               ;; With #+immobile-symbols, all interned symbols are in immobile-space.
-               #+immobile-symbols (sb-xc:symbol-package value)
+               ;; With either of these two features, all interned symbols are
+               ;; as-if static
+               #+(or permgen immobile-symbols) (sb-xc:symbol-package value)
                #-sb-xc-host
                (if (immobile-space-obj-p value)
                    (or (= (generation-of value) +pseudo-static-generation+)
@@ -478,15 +472,15 @@
        immediate-sc-number))
     #+compact-instance-header (layout immediate-sc-number)
     (single-float
-       (if (eql value $0f0) fp-single-zero-sc-number fp-single-immediate-sc-number))
+       (if (eql value 0f0) fp-single-zero-sc-number fp-single-immediate-sc-number))
     (double-float
-       (if (eql value $0d0) fp-double-zero-sc-number fp-double-immediate-sc-number))
+       (if (eql value 0d0) fp-double-zero-sc-number fp-double-immediate-sc-number))
     ((complex single-float)
-       (if (eql value (complex $0f0 $0f0))
+       (if (eql value #c(0f0 0f0))
             fp-complex-single-zero-sc-number
             fp-complex-single-immediate-sc-number))
     ((complex double-float)
-       (if (eql value (complex $0d0 $0d0))
+       (if (eql value #c(0d0 0d0))
             fp-complex-double-zero-sc-number
             fp-complex-double-immediate-sc-number))
     ;; This case has to follow the numeric cases because proxy floating-point numbers
@@ -589,55 +583,6 @@
 (defconstant nargs-offset rcx-offset)
 (defconstant cfp-offset rbp-offset) ; pfw - needed by stuff in /code
 
-(defun combination-implementation-style (node)
-  (declare (type sb-c::combination node))
-  (flet ((valid-funtype (args result)
-           (sb-c::valid-fun-use node
-                                (sb-c::specifier-type
-                                 `(function ,args ,result)))))
-    (case (sb-c::combination-fun-source-name node)
-      (logtest
-       (cond
-         ((or (valid-funtype '(fixnum fixnum) '*)
-              ;; todo: nothing prevents this from testing an unsigned word against
-              ;; a signed word, except for the mess of VOPs it would demand
-              (valid-funtype '((signed-byte 64) (signed-byte 64)) '*)
-              (valid-funtype '((unsigned-byte 64) (unsigned-byte 64)) '*))
-          (values :maybe nil))
-         (t
-          (values :default nil))))
-      (logbitp
-       (if (or
-            (valid-funtype '((mod 64) word) '*)
-            (valid-funtype '((mod 64) signed-word) '*))
-           (values :direct nil)
-           (values :default nil)))
-      (truncate
-       (destructuring-bind (n &optional d) (sb-c::basic-combination-args node)
-         (if (and d
-                  (constant-lvar-p d)
-                  (power-of-two-p (lvar-value d))
-                  (and (csubtypep (sb-c::lvar-type n) (specifier-type 'signed-word))
-                       (not (csubtypep (sb-c::lvar-type n) (specifier-type 'word)))))
-             (values :maybe nil)
-             (values :default nil))))
-      (%dpb
-       (flet ((validp (type result-type)
-                (valid-funtype `((constant-arg integer)
-                                 (constant-arg (integer 1 1))
-                                 ,type
-                                 ,type)
-                               result-type)))
-         (if (and (or (validp 'signed-word 'signed-word)
-                      (validp 'word 'word))
-                  (not (destructuring-bind (new size posn integer) (sb-c::basic-combination-args node)
-                         (declare (ignore new size integer))
-                         (constant-lvar-p posn))))
-             (values :direct nil)
-             (values :default nil))))
-      (t
-       (values :default nil)))))
-
 (defvar *register-names* +qword-register-names+)
 
 ;;; See WRITE-FUNINSTANCE-PROLOGUE in x86-64-vm.
@@ -663,3 +608,28 @@
 #+nil
 (define-cond-sc 32-bit-immediate immediate
   (typep (tn-value tn) '(signed-byte 32)))
+
+(defmacro sb-impl::symbol-allocator-macro (kind name)
+  (declare (ignorable kind))
+  #+permgen
+  `(truly-the symbol (if (eql ,kind 0) ; uninterned
+                         (sb-vm::%alloc-symbol ,name)
+                         (allocate-permgen-symbol ,name)))
+  #-permgen
+  `(truly-the symbol
+          ;; If no immobile-space, easy: all symbols go in dynamic-space
+          #-immobile-space (sb-vm::%alloc-symbol ,name)
+          ;; If #+immobile-symbols, then uninterned symbols go in dynamic space, but
+          ;; interned symbols go in immobile space. Good luck IMPORTing an uninterned symbol-
+          ;; it'll work at least superficially, but if used as a code constant, the symbol's
+          ;; address may violate the assumption that it's an imm32 operand.
+          #+immobile-symbols
+          (if (eql ,kind 0) (sb-vm::%alloc-symbol ,name) (sb-vm::%alloc-immobile-symbol ,name))
+          #+(and immobile-space (not immobile-symbols))
+          (if (or (eql ,kind 1) ; keyword
+                  (and (eql ,kind 2) ; random interned symbol
+                       (plusp (length ,name))
+                       (char= (char ,name 0) #\*)
+                       (char= (char ,name (1- (length ,name))) #\*)))
+              (sb-vm::%alloc-immobile-symbol ,name)
+              (sb-vm::%alloc-symbol ,name))))
