@@ -273,16 +273,6 @@ extern int arch_prctl(int code, unsigned long *addr);
 #define ASSOCIATE_OS_THREAD(thread) thread->os_thread = thread_self()
 #endif
 
-#ifndef LISP_FEATURE_SB_THREAD
-# define ASSIGN_CURRENT_THREAD(dummy)
-#elif defined LISP_FEATURE_GCC_TLS
-# define ASSIGN_CURRENT_THREAD(x) current_thread = x
-#elif !defined LISP_FEATURE_WIN32
-# define ASSIGN_CURRENT_THREAD(x) pthread_setspecific(current_thread, x)
-#else
-# define ASSIGN_CURRENT_THREAD(x) TlsSetValue(OUR_TLS_INDEX, x)
-#endif
-
 #ifdef LISP_FEATURE_WIN32
 // Need a function callable from assembly code, where the inline one won't do.
 void* read_current_thread() {
@@ -404,6 +394,19 @@ void create_main_lisp_thread(lispobj function) {
     ASSIGN_CURRENT_THREAD(NULL);
 }
 
+void sb_posix_after_fork() { // for use by sb-posix:fork
+    struct thread* th = get_sb_vm_thread();
+    th->os_kernel_tid = get_nonzero_tid();
+#ifdef LISP_FEATURE_DARWIN
+    extern void darwin_reinit();
+    darwin_reinit();
+#endif
+#ifdef LISP_FEATURE_MARK_REGION_GC
+    extern void thread_pool_init();
+    thread_pool_init();
+#endif
+}
+
 #ifdef LISP_FEATURE_SB_THREAD
 
 void free_thread_struct(struct thread *th)
@@ -461,11 +464,27 @@ init_new_thread(struct thread *th,
 #endif
 }
 
+lispobj remset_transfer_list;
+
 static void
 unregister_thread(struct thread *th,
                   init_thread_data __attribute__((unused)) *scribble)
 {
     block_blockable_signals(0);
+#ifdef LISP_FEATURE_PERMGEN
+    lispobj my_remset = th->remset;
+    if (my_remset) {
+        lispobj tail = remset_transfer_list;
+        while (1) {
+            VECTOR(my_remset)->data[1] = tail;
+            lispobj actual_old = __sync_val_compare_and_swap(
+                &remset_transfer_list, tail, my_remset);
+            if (actual_old == tail) break;
+            tail = actual_old;
+        }
+        th->remset = 0;
+    }
+#endif
     gc_close_thread_regions(th, LOCK_PAGE_TABLE|CONSUME_REMAINDER);
 #ifdef LISP_FEATURE_SB_SAFEPOINT
     pop_gcing_safety(&scribble->safety);
@@ -544,14 +563,6 @@ void* new_thread_trampoline(void* arg)
     struct thread_instance *lispthread = (void*)native_pointer(th->lisp_thread);
     if (lispthread->_ephemeral_p == LISP_T) th->state_word.user_thread_p = 0;
 
-#ifdef ATOMIC_LOGGING
-      char* string = strdup((char*)VECTOR(name)->data); // FIXME: no such var as 'name'
-      int index = __sync_fetch_and_add(&thread_name_map_count, 1);
-      gc_assert(index < THREAD_NAME_MAP_MAX);
-      thread_name_map[index].thread = pthread_self();
-      thread_name_map[index].name = string;
-#endif
-
     struct vector* startup_info = VECTOR(lispthread->startup_info); // 'lispthread' is pinned
     gc_assert(header_widetag(startup_info->header) == SIMPLE_VECTOR_WIDETAG);
     lispobj startfun = startup_info->data[0]; // 'startup_info' is pinned
@@ -623,6 +634,13 @@ void sb_set_os_thread_name(char* name)
 #if defined LISP_FEATURE_DARWIN && !defined LISP_FEATURE_AVOID_PTHREAD_SETNAME_NP
     if (vector_len(v) < 64) pthread_setname_np(name);
 #endif
+#ifdef ATOMIC_LOGGING
+      char* string = strdup(name);
+      int index = __sync_fetch_and_add(&thread_name_map_count, 1);
+      gc_assert(index < THREAD_NAME_MAP_MAX);
+      thread_name_map[index].thread = pthread_self();
+      thread_name_map[index].name = string;
+#endif
 }
 
 #ifdef LISP_FEATURE_OS_THREAD_STACK
@@ -942,8 +960,9 @@ alloc_thread_struct(void* spaces) {
 #ifdef THREAD_T_NIL_CONSTANTS_SLOT
     tls[THREAD_T_NIL_CONSTANTS_SLOT] = (NIL << 32) | LISP_T;
 #endif
-#ifdef THREAD_ALIEN_LINKAGE_TABLE_BASE_SLOT
-    tls[THREAD_ALIEN_LINKAGE_TABLE_BASE_SLOT] = (lispobj)ALIEN_LINKAGE_TABLE_SPACE_START;
+#if defined LISP_FEATURE_LINKAGE_SPACE && defined LISP_FEATURE_X86_64
+    tls[THREAD_LINKAGE_TABLE_SLOT] = (lispobj)linkage_space;
+    tls[THREAD_ALIEN_LINKAGE_TABLE_BASE_SLOT] = (lispobj)ALIEN_LINKAGE_SPACE_START;
 #endif
 #if defined LISP_FEATURE_X86_64 && defined LISP_FEATURE_LINUX
     tls[THREAD_MSAN_XOR_CONSTANT_SLOT] = 0x500000000000;
@@ -1021,15 +1040,11 @@ alloc_thread_struct(void* spaces) {
     th->state_word.user_thread_p = 1;
 
     lispobj* alien_stack_end = (lispobj*)((char*)th->alien_stack_start + ALIEN_STACK_SIZE);
-#ifdef ALIEN_STACK_GROWS_UPWARD
-    th->alien_stack_pointer = alien_stack_start;
-#elif defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64
-    /* It is not necessary to subtract a word from the starting stack pointer
-     * because consuming C stack space pre-decrements before using the pointer.
-     * So this choice has nothing to do with saving a word of memory, and everything
-     * to do with my lack of understanding of the conventions for non-x86. */
+#if defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64
+    // Alien-stack-pointer is predecremented upon use
     th->alien_stack_pointer = alien_stack_end;
 #else
+    // I do not know the convention for alien-stack-pointer
     th->alien_stack_pointer = alien_stack_end - 1;
 #endif
 
@@ -1109,7 +1124,7 @@ alloc_thread_struct(void* spaces) {
 }
 #ifdef LISP_FEATURE_SB_THREAD
 #ifdef LISP_FEATURE_WIN32
-uword_t create_thread(struct thread* th)
+uword_t create_lisp_thread(struct thread* th)
 {
     unsigned int tid;
     struct extra_thread_data *data = thread_extra_data(th);
@@ -1125,18 +1140,33 @@ uword_t create_thread(struct thread* th)
     }
     return success;
 }
+#else
+#ifdef LISP_FEATURE_OS_THREAD_STACK
+# define START_ROUTINE new_thread_trampoline_switch_stack
+#else
+# define START_ROUTINE new_thread_trampoline
+#endif
+int create_lisp_thread(pthread_t *th, pthread_attr_t *attr, void *arg) {
+    return pthread_create(th, attr, START_ROUTINE, arg);
+}
 #endif
 
 int try_acquire_gc_lock() { return TryEnterCriticalSection(&in_gc_lock); }
 int release_gc_lock() { return mutex_release(&in_gc_lock); }
 
-static __attribute__((unused)) struct timespec stw_begin_time;
-long timespec_delta_microsec(struct timespec* begin, struct timespec* end)
+static __attribute__((unused)) struct timespec stw_begin_realtime, stw_begin_cputime;
+long timespec_diff(struct timespec* begin, struct timespec* end)
 {
+#ifdef LISP_FEATURE_64_BIT
+    return (end->tv_sec - begin->tv_sec) * 1000000000L + (end->tv_nsec - begin->tv_nsec) ;
+#else
     return (end->tv_sec - begin->tv_sec) * 1000000L + (end->tv_nsec - begin->tv_nsec) / 1000;
+#endif
 }
 #ifdef MEASURE_STOP_THE_WORLD_PAUSE
-void thread_accrue_stw_time(struct thread* th, struct timespec* begin)
+void thread_accrue_stw_time(struct thread* th,
+                            struct timespec* begin_real,
+                            struct timespec* begin_cpu)
 {
     /* A non-Lisp thread calling into Lisp via DEFINE-ALIEN-CALLABLE
      * can receive SIG_STOP_FOR_GC as soon as it has a 'struct thread'
@@ -1144,15 +1174,33 @@ void thread_accrue_stw_time(struct thread* th, struct timespec* begin)
     if (th->lisp_thread) {
         struct timespec now;
         clock_gettime(CLOCK_MONOTONIC, &now);
-        unsigned long elapsed = timespec_delta_microsec(begin, &now);
+        unsigned long elapsed = timespec_diff(begin_real, &now);
         struct thread_instance* ti = (void*)INSTANCE(th->lisp_thread);
         if (elapsed > ti->uw_max_stw_pause) ti->uw_max_stw_pause = elapsed;
         ti->uw_sum_stw_pause += elapsed;
         ++ti->uw_ct_stw_pauses;
+        if (begin_cpu) {
+#ifdef CLOCK_THREAD_CPUTIME_ID
+          clock_gettime(CLOCK_THREAD_CPUTIME_ID, &now);
+          ti->uw_gc_virtual_time += timespec_diff(begin_cpu, &now);
+#endif
+        }
     }
 }
 #endif
 
+#ifndef LISP_FEATURE_WIN32
+/* pthread_kill is not guaranteed to be reentrant, prevent
+ * gc_stop_the_world from interrupting another pthread_kill */
+int sb_thread_kill (pthread_t thread, int sig) {
+    sigset_t old;
+    block_blockable_signals(&old);
+    int ret = pthread_kill(thread, sig);
+    thread_sigmask(SIG_SETMASK, &old, NULL);
+    return ret;
+}
+#endif
+
 /* stopping the world is a two-stage process.  From this thread we signal
  * all the others with SIG_STOP_FOR_GC.  The handler for this signal does
  * the usual pseudo-atomic checks (we don't want to stop a thread while
@@ -1193,7 +1241,10 @@ void gc_stop_the_world()
     /* The thread performing stop-the-world does not use sig_stop_for_gc_handler on itself,
      * so it would not accrue time spent stopped. Force it to, by considering it "paused"
      * from the moment it wants to stop all other threads. */
-    clock_gettime(CLOCK_MONOTONIC, &stw_begin_time);
+    clock_gettime(CLOCK_MONOTONIC, &stw_begin_realtime);
+#endif
+#ifdef CLOCK_THREAD_CPUTIME_ID
+    clock_gettime(CLOCK_THREAD_CPUTIME_ID, &stw_begin_cputime);
 #endif
     struct thread *th, *me = get_sb_vm_thread();
     int rc;
@@ -1273,7 +1324,7 @@ void gc_start_the_world()
     lock_ret = mutex_release(&all_threads_lock);
     gc_assert(lock_ret);
 #ifdef MEASURE_STOP_THE_WORLD_PAUSE
-    thread_accrue_stw_time(me, &stw_begin_time);
+    thread_accrue_stw_time(me, &stw_begin_realtime, &stw_begin_cputime);
 #endif
 }
 
